local addonName, a = ...
local L = a.Localize
local s = SpellFlashAddon
local c = BittensGlobalTables.GetTable("BittensSpellFlashLibrary")

local GetComboPoints = GetComboPoints
local UnitInRange = UnitInRange
local UnitIsUnit = UnitIsUnit
local select = select
local math = math

------------------------------------------------------------------------ Common
local function substantialFight()
	return s.Health() > 1.5 * s.MaxHealth("player")
end

local function setFaerieID(z)
	z.Name = nil
	if c.HasTalent("Faerie Swarm") then
		z.ID = c.GetID("Faerie Swarm")
	else
		z.ID = c.GetID("Faerie Fire")
	end
end

c.AddOptionalSpell("Mark of the Wild", nil, {
	Override = function()
		return c.RaidBuffNeeded(c.STAT_BUFFS)
	end
})

c.AddOptionalSpell("Symbiosis", nil, {
	Override = function()
		if c.GetBuffDuration(a.SymbiosisSelfBuffs) > 5 * 60 then
			return false
		end
		for member in c.GetGroupMembers() do
			if not UnitIsUnit(member, "player") 
				and UnitInRange(member) 
				and (s.MyBuff(a.SymbiosisRaidBuffs, member)
					or not s.Buff(a.SymbiosisRaidBuffs, member)) then
				
				return true
			end
		end
	end
})

c.AddOptionalSpell("Faerie Fire", "for Debuff", {
	Debuff = c.ARMOR_DEBUFFS,
	EarlyRefresh = 5,
	RunFirst = setFaerieID,
	CheckFirst = substantialFight,
})

c.AddOptionalSpell("Healing Touch", "Solo", {
	NoRangeCheck = true,
	CheckFirst = function()
		return c.GetHealthPercent("player") < 75
	end
})

c.AddOptionalSpell("Force of Nature", nil, {
	NoRangeCheck = true,
})

c.AddOptionalSpell("Nature's Vigil", nil, {
	NoRangeCheck = true,
})

c.AddSpell("Thrash(Bear Form)", nil, {
	Melee = true,
	NoRangeCheck = true,
})

c.AddInterrupt("Skull Bash", nil, {
	NoGCD = true,
})

----------------------------------------------------------------------- Balance
local function notCastThisEclipse(name)
	return a.EclipsePending
		or c.GetMyDebuffDuration(name) < c.GetBuffDuration("Nature's Grace") - 2
end

c.AddOptionalSpell("Moonkin Form", nil, {
	Type = "form",
})

c.AddOptionalSpell("Incarnation: Chosen of Elune", nil, {
	CheckFirst = function()
		return a.Energy >= 100 or a.Energy <= -100
	end
})

c.AddOptionalSpell("Celestial Alignment", nil, {
	CheckFirst = function()
		return not (a.Lunar or a.Solar)
			and c.GetCooldown("Incarnation: Chosen of Elune") > 10
	end
})

c.AddOptionalSpell("Starfall", nil, {
	Override = function(z)
		if c.GetCooldown("Incarnation: Chosen of Elune") > 15 then
			z.FlashSize = nil
		elseif not a.GoingUp
			and (a.Energy <= -70 
				or (c.GetCooldown("Starsurge") == 0 and a.Energy <= -60)) then
			z.FlashSize = s.FlashSizePercent() / 2
		else
			return false
		end
		return not c.HasBuff("Starfall")
			and (c.GetCooldown("Starfall") == 0
				or (a.EclipsePending and a.Lunar))
	end
})

c.AddSpell("Starsurge", nil, {
	NotIfActive = true,
})

c.AddSpell("Starsurge", "under Shooting Stars", {
	CheckFirst = function()
		return c.HasBuff("Shooting Stars")
	end
})

c.AddSpell("Moonfire", nil, {
	CheckFirst = function()
		return notCastThisEclipse("Moonfire")
	end
})

c.AddSpell("Moonfire", "under Eclipse", {
	EarlyRefresh = 99,
	CheckFirst = function(z)
		return a.Lunar 
			and (notCastThisEclipse("Moonfire") 
				or c.GetMyDebuffDuration("Moonfire") < z.EarlyRefresh)
	end
})
c.ManageDotRefresh("Moonfire under Eclipse", 1, "Moonfire")

c.AddSpell("Sunfire", nil, {
	CheckFirst = function()
		return notCastThisEclipse("Sunfire")
	end
})

c.AddSpell("Sunfire", "under Eclipse", {
	EarlyRefresh = 99,
	CheckFirst = function(z)
		return a.Solar 
			and (notCastThisEclipse("Sunfire")
				or c.GetMyDebuffDuration("Sunfire") < z.EarlyRefresh)
	end
})
c.ManageDotRefresh("Sunfire under Eclipse", 1, "Sunfire")

c.AddSpell("Wrath", "under Celestial Alignment", {
	CheckFirst = function()
		return c.GetBuffDuration("Celestial Alignment") 	
			> c.GetCastTime("Wrath") + .1
	end
})

c.AddSpell("Starfire", nil, {
	CheckFirst = function()
		return a.GoingUp
	end
})

c.AddSpell("Starfire", "under Celestial Alignment", {
	CheckFirst = function()
		return c.GetBuffDuration("Celestial Alignment") 	
			> c.GetCastTime("Starfire") + .1
	end
})

c.AddInterrupt("Solar Beam")

------------------------------------------------------------------------- Feral
local function getDelay(cost, usesCC)
	if usesCC and a.Clearcasting then
		return 0
	end
	
	if a.Berserk > 0 then
		local delay = (cost / 2 - a.Energy) / a.Regen
		if delay < a.Berserk then
			return math.max(0, delay)
		end
	end
	
	return math.max(0, (cost - a.Energy) / a.Regen)
end

local function modForDelay(spell, cost, usesCC)
	local delay = getDelay(cost, usesCC)
	if delay == 0 then
		spell.FlashColor = nil
		spell.FlashSize = nil
	else
		spell.FlashColor = "green"
		spell.FlashSize = s.FlashSizePercent() / 2
	end
	return delay
end

c.AddSpell("Survival Instincts", "under 30", {
	NoGCD = true,
	FlashColor = "red",
	CheckFirst = function()
		return c.GetHealthPercent("player") < 30
	end
})

c.AddSpell("Barkskin", "under 30", {
	NoGCD = true,
	FlashColor = "red",
	CheckFirst = function()
		return c.GetHealthPercent("player") < 30
	end
})

c.AddOptionalSpell("Maul", "for Feral", {
	Melee = true,
	Override = function()
		return a.Rage >= 30 and c.GetCooldown("Maul", true) == 0
	end,
})

c.AddSpell("Mangle(Bear Form)", "for Feral", {
	CheckFirst = function()
		return (100 - a.Energy) / a.Regen > 2
	end
})

c.AddSpell("Cat Form", nil, {
	Type = "form"
})

c.AddOptionalSpell("Tiger's Fury", nil, {
	NoGCD = true,
	CheckFirst = function()
		return a.Energy <= 35 
			and not a.Clearcasting
			and not c.HasBuff("Berserk", true)
	end
})

c.AddOptionalSpell("Berserk", nil, {
	NoGCD = true,
	CheckFirst = function(z)
		local furyCool = c.GetCooldown("Tiger's Fury")
		if furyCool > 6 or not substantialFight() then
			z.FlashColor = "yellow"
		else
			z.FlashColor = "green"
		end
		
		if furyCool == 0 and substantialFight() then
			z.FlashSize = s.FlashSizePercent() / 2
		else
			z.FlashSize = nil
		end
		
		return furyCool > 6 or furyCool == 0 or not substantialFight()
	end
})

c.AddOptionalSpell("Incarnation: King of the Jungle")

c.AddOptionalSpell("Healing Touch", "for Feral", {
	Override = function()
		return a.DreamStacks <= 0
			and a.Swiftness
			and c.HasTalent("Dream of Cenarius")
	end
})

c.AddOptionalSpell("Healing Touch", "for Feral Heal", {
	Override = function()
		return c.GetHealthPercent("player") < 80
			and a.Swiftness 
			and not c.HasTalent("Dream of Cenarius")
	end
})

local function setRoarId(spell)
	if c.HasGlyph("Savagery") then
		spell.ID = c.GetID("Savage Roar Glyphed")
	else
		spell.ID = c.GetID("Savage Roar Unglyphed")
	end
end

c.AddSpell("Savage Roar", nil, {
	RunFirst = setRoarId,
	Override = function(z)
		return (a.CP > 0 or c.HasGlyph("Savagery"))
			and modForDelay(z, 25) >= a.Roar
	end,
})

c.AddSpell("Savage Roar", "Early", {
	RunFirst = setRoarId,
	Override = function(z)
		return a.CP == 5
			and a.Roar < a.Rip + 2
			and a.Roar - modForDelay(z, 25) < 6
			and substantialFight()
	end
})

c.AddSpell("Ferocious Bite", "on Last Tick", {
	Melee = true,
	Override = function(z)
		local rip = a.Rip - modForDelay(z, 25, true)
		return a.InExecute and a.CP > 0 and rip > 0 and rip < 2
	end
})

c.AddSpell("Ferocious Bite", "at 8", {
	Melee = true,
	Override = function(z)
		local delay = modForDelay(z, 25, true)
		return a.CP == 5
			and ((a.InExecute and a.Rip - delay < 8) or not substantialFight())
	end
})

c.AddSpell("Ferocious Bite", "w/ Soul of the Forest", {
	Melee = true,
	Override = function(z)
		return a.CP == 5
			and not a.InExecute
			and c.HasTalent("Soul of the Forest")
			and a.Rip - modForDelay(z, 25, true) > 14
	end
})

c.AddSpell("Thrash(Cat Form)", nil, {
	Melee = true,
	Override = function()
		return a.Clearcasting
			and a.King == 0
			and c.GetMyDebuffDuration("Thrash(Cat Form)", false, true) < 3
			and substantialFight()
	end
})

local function swiftable()
	return not a.Swiftness
		and a.DreamStacks == 0
		and a.CP == 5
		and c.GetCooldown("Nature's Swiftness") == 0
		and c.HasTalent("Dream of Cenarius")
		and substantialFight()
end

c.AddOptionalSpell("Nature's Swiftness", "pre-Rip", {
	Override = function()
		if swiftable() and (not a.InExecute or a.Rip == 0) and a.Rip < 3.5 then
			s.Flash(c.GetID("Healing Touch"), nil, s.FlashSizePercent() / 2)
			return true
		end
	end
})

c.AddOptionalSpell("Nature's Swiftness", "pre-Bite", {
	Override = function()
		if swiftable() and a.InExecute and a.Rip < 9.5 then
			s.Flash(c.GetID("Healing Touch"), nil, s.FlashSizePercent() / 2)
			return true
		end
	end
})

c.AddSpell("Rip", nil, {
	Melee = true,
	Override = function(z)
		return a.CP == 5
			and (not a.InExecute or a.Rip == 0)
			and a.Rip - modForDelay(z, 30, true) < 2
			and substantialFight()
	end
})

c.AddSpell("Rake", nil, {
	Melee = true,
	Override = function(z)
		local dur = c.GetMyDebuffDuration("Rake") - modForDelay(z, 35, true)
		return ((dur < 6 and a.DreamStacks > 0)
				or (dur < 3 and (a.King == 0 or s.HealthPercent() <= 80)))
			and not c.IsAuraPendingFor("Rake")
	end
})

c.AddSpell("Bear Form", "for Feral", {
	Type = "form",
	Melee = true,
	CheckFirst = function()
		return not c.HasMyDebuff("Thrash(Bear Form)", nil, true)
			and (100 - a.Energy) / a.Regen > 2.5
			and a.Berserk == 0
			and not c.IsCasting("Bear Form")
			and substantialFight()
	end
})

c.AddSpell("Ravage", nil, {
	FlashID = { "Ravage", "Ravage!" },
	Melee = true,
	Override = function(z)
		return modForDelay(z, 45, true) < a.King
	end
})

c.AddSpell("Shred", nil, {
	FlashID = { "Shred", "Shred!" },
	Melee = true,
	Override = function(z)
		local delay = modForDelay(z, 40, true)
		return not (a.NoShred[s.UnitInfo()] or c.IsTanking())
			or (c.HasGlyph("Shred") 
				and (a.TigersFury > delay or a.Berserk > delay))
	end
})

c.AddSpell("Mangle(Cat Form)", nil, {
	Melee = true,
	Override = function(z)
		modForDelay(z, 35, true)
		return true
	end
})

---------------------------------------------------------------------- Guardian
c.AddOptionalSpell("Bear Form", nil, {
	Type = "form",
})

c.AddOptionalSpell("Tooth and Claw", nil, {
	NoGCD = true,
	IsUp = function()
		return c.HasMyDebuff("Tooth and Claw", true)
	end,
})

c.AddOptionalSpell("Cenarion Ward", nil, {
	NoRangeCheck = true,
})

c.AddOptionalSpell("Nature's Swiftness", "for Guardian", {
	NoRangeCheck = true,
	InactiveFlashID = { "Healing Touch", "Nature's Swiftness" },
	RunFirst = function(z)
		if c.HasBuff("Nature's Swiftness") then
			z.ID = c.GetID("Healing Touch")
			z.FlashID = nil
		else
			z.ID = c.GetID("Nature's Swiftness")
			z.FlashID = z.InactiveFlashID
		end
	end, 
	ShouldHold = function()
		return s.HealthPercent("player") > 80
	end,
})

c.AddOptionalSpell("Barkskin", nil, {
	NoGCD = true,
})

c.AddOptionalSpell("Renewal", nil, {
	NoGCD = true,
	ShouldHold = function()
		return s.HealthPercent("player") > 70
	end,
})

c.AddOptionalSpell("Might of Ursoc", "at 2 mins", {
	NoGCD = true,
	IsUp = function()
		return c.HasBuff("Might of Ursoc", true)
			and c.WearingSet(2, "GuardianT14")
	end,
	Enabled = function()
		return c.WearingSet(2, "GuardianT14")
			and not c.HasGlyph("Might of Ursoc")
	end,
	ShouldHold = function()
		if c.HasGlyph("Might of Ursoc") then
			return s.HealthPercent("player") > 70
		else
			return s.HealthPercent("player") > 50
		end
	end,
})

c.AddOptionalSpell("Might of Ursoc", "above 2 mins", {
	NoGCD = true,
	Enabled = function()
		return not c.WearingSet(2, "GuardianT14") 
			or c.HasGlyph("Might of Ursoc")
	end,
	ShouldHold = function()
		if c.HasGlyph("Might of Ursoc") then
			return s.HealthPercent("player") > 70
		else
			return s.HealthPercent("player") > 50
		end
	end,
})

c.AddOptionalSpell("Survival Instincts", "Unglyphed", {
	NoGCD = true,
	Enabled = function()
		return not c.HasGlyph("Survival Instincts")
	end,
	CheckFirst = function()
		return not c.HasBuff("Frenzied Regeneration", true)
	end,
})

c.AddOptionalSpell("Survival Instincts", "Glyphed", {
	NoGCD = true,
	Enabled = function()
		return c.HasGlyph("Survival Instincts")
	end,
	CheckFirst = function()
		return not c.HasBuff("Frenzied Regeneration", true)
	end,
})

c.AddOptionalSpell("Frenzied Regeneration", nil, {
	NoGCD = true,
	CheckFirst = function()
		if c.HasBuff("Savage Defense", true) or not c.IsTanking() then
			return false
		end
		
		local damage = s.HealthDamagePercent("player")
		if c.HasGlyph("Frenzied Regeneration") then
			return damage > 15 and a.Rage >= 60
		else
			return damage > math.min(60, a.Rage) * .3
		end
	end,
})

c.AddOptionalSpell("Savage Defense", nil, {
	NoGCD = true,
	Melee = true,
	CheckFirst = function()
		return not c.HasBuff("Frenzied Regeneration") and c.IsTanking()
	end
})

c.AddOptionalSpell("Maul", nil, {
	NoGCD = true,
	CheckFirst = function()
		return a.EmptyRage < 10 and c.HasBuff("Tooth and Claw", true)
	end,
})

c.AddTaunt("Growl", nil, {
	NoGCD = true,
})

c.AddSpell("Thrash(Bear Form)", "for Weakened Blows", {
	Melee = true,
	NoRangeCheck = true,
	CheckFirst = function()
		return not c.HasDebuff(c.WEAKENED_BLOWS_DEBUFFS)
	end,
})

c.AddSpell("Thrash(Bear Form)", "for Bleed", {
	Melee = true,
	NoRangeCheck = true,
	MyDebuff = "Thrash(Bear Form)",
	EarlyRefresh = 2,
})

c.AddOptionalSpell("Healing Touch", "for Guardian", {
	CheckFirst = function()
		return c.HasBuff("Nature's Swiftness") 
			and s.HealthPercent("player") < 80
	end,
})

c.AddOptionalSpell("Rejuvenation", "for Guardian", {
	CheckFirst = function(z)
		return c.HasBuff("Heart of the Wild") and not c.HasBuff("Rejuvenation")
	end,
})

c.AddOptionalSpell("Rejuvenation", "Refresh for Guardian", {
	EarlyRefresh = 99,
	CheckFirst = function(z)
		return c.HasBuff("Heart of the Wild")
			and c.GetBuffDuration("Rejuvenation") < z.EarlyRefresh
	end,
})
c.ManageDotRefresh("Rejuvenation Refresh for Guardian", 3, "Rejuvenation")

c.AddSpell("Faerie Fire", nil, {
	RunFirst = setFaerieID,
})

c.AddSpell("Faerie Fire", "for Debuff for Guardian", {
	Debuff = c.ARMOR_DEBUFFS,
	RunFirst = setFaerieID,
})

c.AddSpell("Enrage", nil, {
	NoGCD = true,
	CheckFirst = function()
		return a.EmptyRage > 20 and c.IsTanking()
	end,
})
